// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "sequence.h"
#include <iostream>

// Load NCBI headers
#include "objseq.h"

using namespace std;

// Should we wrap output?
bool wrap_seq_output = false;

// Local functions for 3 bit DNA base encoding
inline unsigned char CharToBase3(const unsigned char &c);
inline unsigned char CharToComplementBase3(const unsigned char &c);
inline unsigned char BaseToChar3(const unsigned char &c);

inline void WriteUnit3(unsigned char unit[3], unsigned char *ptr);
inline void WriteUnit3Remainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder);

inline void WriteUnit3Complement(unsigned char unit[3], unsigned char *ptr);
inline void WriteUnit3ComplementRemainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder);

// Local functions for 2 bit DNA base encoding
inline unsigned char CharToBase2(const unsigned char &c);
inline unsigned char CharToComplementBase2(const unsigned char &c);
inline unsigned char BaseToChar2(const unsigned char &c);

inline void WriteUnit2(unsigned char *unit, unsigned char *ptr);
inline void WriteUnit2Remainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder);

inline void WriteUnit2Complement(unsigned char *unit, unsigned char *ptr);
inline void WriteUnit2ComplementRemainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder);

/////////////////////////////////////////////////////////////////////
// Code for 3 bit DNA encoding
/////////////////////////////////////////////////////////////////////
	
DNA3Seq::DNA3Seq()
{
	seq = NULL;
	len = 0;
	num_base = 0;
}

DNA3Seq::DNA3Seq(const std::string &m_seq)
{
	seq = NULL;
	len = 0;
	num_base = 0;
	
	*this = m_seq;
}

DNA3Seq::DNA3Seq(const std::string &m_seq, const int &m_from, const int &m_len,
		const char &m_strand)
{
	seq = NULL;
	len = 0;
	num_base = 0;

	if(m_strand == Seq_strand_plus){
		*this = m_seq.substr(m_from, m_len);
	}
	else{ // m_strand == Seq_strand_minus
		complement(m_seq.substr(m_from, m_len));
	}
}

DNA3Seq::DNA3Seq(const DNA3Seq &m_seq, const int &m_from, const int &m_len,
		const char &m_strand)
{
	seq = NULL;
	len = 0;
	num_base = 0;

	if(m_strand == Seq_strand_plus){
		*this = m_seq.subseq(m_from, m_len);
	}
	else{ // m_strand == Seq_strand_minus
		complement(m_seq.subseq(m_from, m_len));
	}
}

DNA3Seq::DNA3Seq(const char *m_seq)
{
	seq = NULL;
	len = 0;
	num_base = 0;
	
	*this = m_seq;
}

DNA3Seq::DNA3Seq(const DNA3Seq &m_seq)
{
	num_base = m_seq.num_base;
	len = m_seq.len;
	
	if(len == 0){
		seq = NULL;
	}
	else{
		unsigned int tmp = len*DNA3UnitSizeInByte;
		
		seq = new unsigned char[tmp];
		
		memcpy(seq, m_seq.seq, tmp);
	}
}

DNA3Seq::DNA3Seq(const DNA2Seq &m_seq)
{
	seq = NULL;
	*this = m_seq;
}

DNA3Seq::DNA3Seq(const int &m_size, const unsigned char &m_default_base /* = NONE*/)
{
	// m_size is the number of bases
	
	// How much space will be required to store the sequence?
	num_base = m_size;
	
	unsigned int seq_units = num_base/DNA3UnitSizeInBase;
	unsigned int seq_remainder = num_base%DNA3UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	// Initialize all memory to m_default_base
	unsigned int i;
	unsigned char *ptr;
	
	for(i = 0, ptr = seq;i < len;i++, ptr += DNA3UnitSizeInByte){
		ptr[0] =  (unsigned char)(m_default_base << 5) + 
				(unsigned char)(m_default_base << 2) + 
				(unsigned char)(m_default_base >> 1);
		ptr[1] =  (unsigned char)(m_default_base << 7) + 
				(unsigned char)(m_default_base << 4) + 
				(unsigned char)(m_default_base << 1) + 
				(unsigned char)(m_default_base >> 2);
		ptr[2] =  (unsigned char)(m_default_base << 6) + 
				(unsigned char)(m_default_base << 3) + 
				(unsigned char)m_default_base;
	}
}
	
DNA3Seq::~DNA3Seq()
{
	if(seq){
		delete [] seq;
	}
}

DNA3Seq& DNA3Seq::operator=(const DNA3Seq &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	num_base = m_seq.num_base;
	len = m_seq.len;
	
	if(len == 0){
		seq = NULL;
	}
	else{
		unsigned int tmp = len*DNA3UnitSizeInByte;
		
		seq = new unsigned char[tmp];
		
		memcpy(seq, m_seq.seq, tmp);
	}
	
	return (*this);
}

DNA3Seq& DNA3Seq::operator=(const DNA2Seq &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	num_base = m_seq.num_base;
	
	unsigned int seq_units = num_base/DNA3UnitSizeInBase;
	unsigned int seq_remainder = num_base%DNA3UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	iterator iter = begin();
	DNA2Seq::iterator copy_iter = m_seq.begin();
	
	while(copy_iter.is_valid()){
		iter.base(copy_iter.base());
		
		copy_iter++;
		iter++;
	}
	
	return (*this);
}

DNA3Seq& DNA3Seq::operator=(const string &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = m_seq.size();
	unsigned int seq_units = num_base/DNA3UnitSizeInBase;
	unsigned int seq_remainder = num_base%DNA3UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return (*this);
	}
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq.c_str();
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr += DNA3UnitSizeInBase, head += DNA3UnitSizeInByte){
		WriteUnit3(head, ptr);
	}
	
	WriteUnit3Remainder(head, ptr, seq_remainder);
		
	return (*this);
}

DNA3Seq& DNA3Seq::operator=(const char *m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = strlen(m_seq);
	unsigned int seq_units = num_base/DNA3UnitSizeInBase;
	unsigned int seq_remainder = num_base % DNA3UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return (*this);
	}
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq;
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr += DNA3UnitSizeInBase, head += DNA3UnitSizeInByte){
				
		WriteUnit3(head, ptr);
	}
	
	WriteUnit3Remainder(head, ptr, seq_remainder);
		
	return (*this);
}

DNA3Seq DNA3Seq::operator+(const DNA3Seq &m_seq)
{	
	DNA3Seq tmp(this->size() + m_seq.size());
	
	DNA3Seq::iterator iter = tmp.begin();
	
	DNA3Seq::iterator copy_iter;
	
	for(copy_iter = this->begin();copy_iter != this->end();copy_iter++, iter++){
		iter.base( copy_iter.base() );
	}
	
	for(copy_iter = m_seq.begin();copy_iter != m_seq.end();copy_iter++, iter++){
		iter.base( copy_iter.base() );
	}
	
	return tmp;
}

DNA3Seq DNA3Seq::subseq(const unsigned int &m_start, const unsigned int &m_len,
	const bool &m_circular /* = false */) const
{
	iterator iter;
	
	if(m_circular){
		iter = begin() + m_start % size();
	}
	else{
		if((m_start + m_len - 1) >= size()){
			throw __FILE__ ":DNA3Seq::subseq: Illegal range";
		}
		
		iter = begin() + m_start;
	}
	
	DNA3Seq tmp(m_len);
	iterator tmp_iter = tmp.begin();
	
	// This was the original version that did not handle periodic sequences
	//iterator iter = begin() + m_start;
	
	for(unsigned int i = 0;i < m_len;i++, iter++, tmp_iter++){
		if(m_circular && (iter == end())){
			iter = begin();
		}
		
		tmp_iter.base( iter.base() );
	}
	
	return tmp;
}

// Make a copy of the complement of the given sequence
// where complement is the reversed sequence with following
// base substitutions:
// A->T'
// T->A'
// C->G'
// G->C'
void DNA3Seq::complement(const char *m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = strlen(m_seq);
	unsigned int seq_units = num_base/DNA3UnitSizeInBase;
	unsigned int seq_remainder = num_base % DNA3UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq + num_base - 1;
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr -= DNA3UnitSizeInBase, head += DNA3UnitSizeInByte){
				
		WriteUnit3Complement(head, ptr);
	}
	
	WriteUnit3ComplementRemainder(head, ptr, seq_remainder);
}

// Make a copy of the complement of the given sequence
// where complement is the reversed sequence with following
// base substitutions:
// A->T'
// T->A'
// C->G'
// G->C'
void DNA3Seq::complement(const string &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = m_seq.size();
	unsigned int seq_units = num_base/DNA3UnitSizeInBase;
	unsigned int seq_remainder = num_base % DNA3UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq.c_str() + num_base - 1;
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr -= DNA3UnitSizeInBase, head += DNA3UnitSizeInByte){
				
		WriteUnit3Complement(head, ptr);
	}
	
	WriteUnit3ComplementRemainder(head, ptr, seq_remainder);
}

void DNA3Seq::complement(const DNA3Seq &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = m_seq.num_base;
	len = m_seq.len;
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len*DNA3UnitSizeInByte];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA3Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	// Set the entire last unit to have NONE for all bases. This
	// will insure that we have the correct remainder after filling
	// in only the valid bases.
	unsigned char *ptr = seq + (len - 1)*DNA3UnitSizeInByte;
	
	ptr[0] =  (unsigned char)(DNA3Seq::NONE << 5) + 
			(unsigned char)(DNA3Seq::NONE << 2) + 
			(unsigned char)(DNA3Seq::NONE >> 1);
	ptr[1] =  (unsigned char)(DNA3Seq::NONE << 7) + 
			(unsigned char)(DNA3Seq::NONE << 4) + 
			(unsigned char)(DNA3Seq::NONE << 1) + 
			(unsigned char)(DNA3Seq::NONE >> 2);
	ptr[2] =  (unsigned char)(DNA3Seq::NONE << 6) + 
			(unsigned char)(DNA3Seq::NONE << 3) + 
			(unsigned char)DNA3Seq::NONE;
	
	complement_iterator comp_iter = m_seq.complement_begin();
	iterator iter = begin();

	while(comp_iter != m_seq.complement_end()){
		iter.base(comp_iter.base());
		
		iter++;
		comp_iter++;
	}
}

DNA3Seq DNA3Seq::complement() const
{
	DNA3Seq tmp;

	tmp.complement(*this);

	return tmp;
}
	
// Convert this DNA3Seq into a text string
string  DNA3Seq::str() const
{
	string tmp;
	int i;
	
	tmp.resize(num_base);
	
	DNA3Seq::iterator iter;
	
	for(i = 0, iter =  begin();iter !=  end();iter++, i++){
		tmp[i] = iter.c_base();
	}
	
	return tmp;
}

bool DNA3Seq::is_ATGC() const
{
	DNA3Seq::iterator iter;
	
	for(iter = begin();iter != end();iter++){
		// Assume that T is the largest valid base
		if(iter.base() > DNA3Seq::T){
			return false;
		}
	}

	return true;
}

float DNA3Seq::gc_content() const
{
	DNA3Seq::iterator iter;
	float local_gc = 0.0f;

	for(iter = begin();iter != end();iter++){
		const unsigned char b = iter.base();

		local_gc += ( (b == DNA3Seq::G) || (b == DNA3Seq::C) );
	}

	return empty() ? 0.0f : local_gc/num_base;
}

ostream& operator << (ostream &s, const DNA3Seq &m_seq)
{
	DNA3Seq::iterator iter;
	unsigned int count = 0;
	
	for(iter =  m_seq.begin();iter !=  m_seq.end();iter++){
		
		s << iter.c_base();
		
		if(wrap_seq_output){
			count++;

			if(count % DNA_OUTPUT_LINE_LENGTH == 0){
				s << endl;
			}
		}
	}
	
	return s;
}

inline void WriteUnit3(unsigned char *unit, unsigned char *ptr)
{
	unsigned char tmp2 = CharToBase3(ptr[2]);
	unsigned char tmp5 = CharToBase3(ptr[5]);
	
	unit[0] = (CharToBase3(ptr[0]) << 5) + 
			(CharToBase3(ptr[1]) << 2) + 
			(tmp2 >> 1);
	unit[1] = (tmp2 << 7) + 
			(CharToBase3(ptr[3]) << 4) + 
			(CharToBase3(ptr[4]) << 1) + 
			(tmp5 >> 2);
	unit[2] = (tmp5 << 6) + 
			(CharToBase3(ptr[6]) << 3) + 
			CharToBase3(ptr[7]);
}

inline void WriteUnit3Remainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder)
{
	unsigned char tmp2;
	unsigned char tmp5;
	
	switch(remainder){
		case 0:
			return;
		case 1:
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(unsigned char)(DNA3Seq::NONE << 2) + 
					(unsigned char)(DNA3Seq::NONE >> 1);
			unit[1] = (unsigned char)(DNA3Seq::NONE << 7) + 
					(unsigned char)(DNA3Seq::NONE << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					(unsigned char)DNA3Seq::NONE;
			
			return;
		case 2:
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(CharToBase3(ptr[1]) << 2) + 
					(unsigned char)(DNA3Seq::NONE >> 1);
			unit[1] = (unsigned char)(DNA3Seq::NONE << 7) + 
					(unsigned char)(DNA3Seq::NONE << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					(unsigned char)DNA3Seq::NONE;
					 
			return;
		case 3:
			tmp2 = CharToBase3(ptr[2]);
			
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(CharToBase3(ptr[1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(unsigned char)(DNA3Seq::NONE << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					(unsigned char)DNA3Seq::NONE;
			return;
		case 4:
			tmp2 = CharToBase3(ptr[2]);
			
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(CharToBase3(ptr[1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToBase3(ptr[3]) << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					(unsigned char)DNA3Seq::NONE;
			return;
		case 5:
			tmp2 = CharToBase3(ptr[2]);
			
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(CharToBase3(ptr[1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToBase3(ptr[3]) << 4) + 
					(CharToBase3(ptr[4]) << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					(unsigned char)DNA3Seq::NONE;
					 
			return;
		case 6:
			tmp2 = CharToBase3(ptr[2]);
			tmp5 = CharToBase3(ptr[5]);
			
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(CharToBase3(ptr[1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToBase3(ptr[3]) << 4) + 
					(CharToBase3(ptr[4]) << 1) + 
					(tmp5 >> 2);
			unit[2] = (tmp5 << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					(unsigned char)DNA3Seq::NONE;
			return;
		case 7:
			tmp2 = CharToBase3(ptr[2]);
			tmp5 = CharToBase3(ptr[5]);
			
			unit[0] = (CharToBase3(ptr[0]) << 5) + 
					(CharToBase3(ptr[1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToBase3(ptr[3]) << 4) + 
					(CharToBase3(ptr[4]) << 1) + 
					(tmp5 >> 2);
			unit[2] = (tmp5 << 6) + 
					(CharToBase3(ptr[6]) << 3) + 
					(unsigned char) DNA3Seq::NONE;
			return;
	};
}

inline void WriteUnit3Complement(unsigned char *unit, unsigned char *ptr)
{
	unsigned char tmp2 = CharToComplementBase3(ptr[-2]);
	unsigned char tmp5 = CharToComplementBase3(ptr[-5]);
	
	unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
			(CharToComplementBase3(ptr[-1]) << 2) + 
			(tmp2 >> 1);
	unit[1] = (tmp2 << 7) + 
			(CharToComplementBase3(ptr[-3]) << 4) + 
			(CharToComplementBase3(ptr[-4]) << 1) + 
			(tmp5 >> 2);
	unit[2] = (tmp5 << 6) + 
			(CharToComplementBase3(ptr[-6]) << 3) + 
			CharToComplementBase3(ptr[-7]);
}

inline void WriteUnit3ComplementRemainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder)
{
	unsigned char tmp2;
	unsigned char tmp5;
	
	switch(remainder){
		case 0:
			return;
		case 1:
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(unsigned char)(DNA3Seq::NONE << 2) + 
					(unsigned char)(DNA3Seq::NONE >> 1);
			unit[1] = (unsigned char)(DNA3Seq::NONE << 7) + 
					(unsigned char)(DNA3Seq::NONE << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					 (unsigned char)DNA3Seq::NONE;

			return;
		case 2:
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(CharToComplementBase3(ptr[-1]) << 2) + 
					(unsigned char)(DNA3Seq::NONE >> 1);
			unit[1] = (unsigned char)(DNA3Seq::NONE << 7) + 
					(unsigned char)(DNA3Seq::NONE << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					 (unsigned char)DNA3Seq::NONE;
			return;
		case 3:
			tmp2 = CharToComplementBase3(ptr[-2]);
			
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(CharToComplementBase3(ptr[-1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(unsigned char)(DNA3Seq::NONE << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					 (unsigned char)DNA3Seq::NONE;
			return;
		case 4:
			tmp2 = CharToComplementBase3(ptr[-2]);
			
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(CharToComplementBase3(ptr[-1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToComplementBase3(ptr[-3]) << 4) + 
					(unsigned char)(DNA3Seq::NONE << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					 (unsigned char)DNA3Seq::NONE;
			return;
		case 5:
			tmp2 = CharToComplementBase3(ptr[-2]);
			
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(CharToComplementBase3(ptr[-1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToComplementBase3(ptr[-3]) << 4) + 
					(CharToComplementBase3(ptr[-4]) << 1) + 
					(unsigned char)(DNA3Seq::NONE >> 2);
			unit[2] = (unsigned char)(DNA3Seq::NONE << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					 (unsigned char)DNA3Seq::NONE;
					 
			return;
		case 6:
			tmp2 = CharToComplementBase3(ptr[-2]);
			tmp5 = CharToComplementBase3(ptr[-5]);
			
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(CharToComplementBase3(ptr[-1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToComplementBase3(ptr[-3]) << 4) + 
					(CharToComplementBase3(ptr[-4]) << 1) + 
					(tmp5 >> 2);
			unit[2] = (tmp5 << 6) + 
					(unsigned char)(DNA3Seq::NONE << 3) + 
					 (unsigned char)DNA3Seq::NONE;
					 
			return;
		case 7:
			tmp2 = CharToComplementBase3(ptr[-2]);
			tmp5 = CharToComplementBase3(ptr[-5]);
			
			unit[0] = (CharToComplementBase3(ptr[0]) << 5) + 
					(CharToComplementBase3(ptr[-1]) << 2) + 
					(tmp2 >> 1);
			unit[1] = (tmp2 << 7) + 
					(CharToComplementBase3(ptr[-3]) << 4) + 
					(CharToComplementBase3(ptr[-4]) << 1) + 
					(tmp5 >> 2);
			unit[2] = (tmp5 << 6) + 
					(CharToComplementBase3(ptr[-6]) << 3) + 
					(unsigned char) DNA3Seq::NONE;
			return;
	};
}

inline unsigned char CharToBase3(const unsigned char &c)
{
	switch(c){
		case 'a':
		case 'A':
			return DNA3Seq::A;
		case 't':
		case 'T':
			return DNA3Seq::T;
		case 'g':
		case 'G':
			return DNA3Seq::G;
		case 'c':
		case 'C':
			return DNA3Seq::C;
		case '-':
			return DNA3Seq::GAP;
	};
	
	return DNA3Seq::UNKNOWN;
}

inline unsigned char CharToComplementBase3(const unsigned char &c)
{
	switch(c){
		case 'a':
		case 'A':
			return DNA3Seq::T;
		case 't':
		case 'T':
			return DNA3Seq::A;
		case 'g':
		case 'G':
			return DNA3Seq::C;
		case 'c':
		case 'C':
			return DNA3Seq::G;
		case '-':
			return DNA3Seq::GAP;
	};
	
	return DNA3Seq::UNKNOWN;
}

inline unsigned char BaseToChar3(const unsigned char &c)
{	
	switch(c & BITMASK3){
		case DNA3Seq::A:
			return 'A';
		case DNA3Seq::T:
			return 'T';
		case  DNA3Seq::G:
			return 'G';
		case DNA3Seq::C:
			return 'C';
		case DNA3Seq::GAP:
			return '-';
		case DNA3Seq::UNKNOWN:
			return 'N';
	};
	
	return DNA3Seq::NONE;
}

/////////////////////////////////////////////////////////////////////
// Code for 2 bit DNA encoding
/////////////////////////////////////////////////////////////////////
	
DNA2Seq::DNA2Seq()
{
	seq = NULL;
	len = 0;
	num_base = 0;
	seq_remainder = 0;
}

DNA2Seq::DNA2Seq(const std::string &m_seq)
{
	seq = NULL;
	len = 0;
	num_base = 0;
	seq_remainder = 0;
	
	*this = m_seq;
}

DNA2Seq::DNA2Seq(const std::string &m_seq, const int &m_from, const int &m_len,
		const char &m_strand)
{
	seq = NULL;
	len = 0;
	num_base = 0;
	seq_remainder = 0;

	if(m_strand == Seq_strand_plus){
		*this = m_seq.substr(m_from, m_len);
	}
	else{ // m_strand == Seq_strand_minus
		complement(m_seq.substr(m_from, m_len));
	}
}

DNA2Seq::DNA2Seq(const DNA2Seq &m_seq, const int &m_from, const int &m_len,
		const char &m_strand)
{
	seq = NULL;
	len = 0;
	num_base = 0;
	seq_remainder = 0;

	if(m_strand == Seq_strand_plus){
		*this = m_seq.subseq(m_from, m_len);
	}
	else{ // m_strand == Seq_strand_minus
		complement(m_seq.subseq(m_from, m_len));
	}
}

DNA2Seq::DNA2Seq(const char *m_seq)
{
	seq = NULL;
	len = 0;
	num_base = 0;
	seq_remainder = 0;
	
	*this = m_seq;
}

DNA2Seq::DNA2Seq(const DNA2Seq &m_seq)
{
	num_base = m_seq.num_base;
	len = m_seq.len;
	seq_remainder = m_seq.seq_remainder;
	
	if(len == 0){
		seq = NULL;
	}
	else{
		unsigned int tmp = len;
		
		seq = new unsigned char[tmp];
		
		memcpy(seq, m_seq.seq, tmp);
	}
}

DNA2Seq::DNA2Seq(const DNA3Seq &m_seq)
{
	seq = NULL;
	
	*this = m_seq;
}

DNA2Seq::DNA2Seq(const int &m_size)
{
	// m_size is the number of bases
	
	// How much space will be required to store the sequence?
	num_base = m_size;
	
	unsigned int seq_units = num_base/DNA2UnitSizeInBase;
	seq_remainder = num_base%DNA2UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len];
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	// Initialize the memory to all A (this is to make sure
	// comparisons between sequences are rational!).
	memset(seq, DEFAULT_BYTE, len);
}

DNA2Seq::~DNA2Seq()
{
	if(seq){
		delete [] seq;
	}
}

DNA2Seq& DNA2Seq::operator=(const DNA2Seq &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	num_base = m_seq.num_base;
	len = m_seq.len;
	seq_remainder = m_seq.seq_remainder;
	
	if(len == 0){
		seq = NULL;
	}
	else{
		unsigned int tmp = len;
		
		seq = new unsigned char[tmp];
		
		memcpy(seq, m_seq.seq, tmp);
	}
	
	return (*this);
}

DNA2Seq& DNA2Seq::operator=(const DNA3Seq &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	num_base = m_seq.num_base;
	
	unsigned int seq_units = num_base/DNA2UnitSizeInBase;
	seq_remainder = num_base%DNA2UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	seq = new unsigned char[len];
	
	// Only need to initialize the last byte
	seq[len - 1] = DEFAULT_BYTE;
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	iterator iter = begin();
	DNA3Seq::iterator copy_iter = m_seq.begin();
	
	while(copy_iter.is_valid()){
		iter.base(copy_iter.base());
		
		copy_iter++;
		iter++;
	}
	
	return (*this);
}

DNA2Seq& DNA2Seq::operator=(const string &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = m_seq.size();
	unsigned int seq_units = num_base/DNA2UnitSizeInBase;
	seq_remainder = num_base%DNA2UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return (*this);
	}
	
	seq = new unsigned char[len];
	
	// Only need to initialize the last byte
	seq[len - 1] = DEFAULT_BYTE;
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq.c_str();
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr += DNA2UnitSizeInBase, head++ /*head += DNA2UnitSizeInByte*/){
		WriteUnit2(head, ptr);
	}
	
	WriteUnit2Remainder(head, ptr, seq_remainder);
		
	return (*this);
}

DNA2Seq& DNA2Seq::operator=(const char *m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = strlen(m_seq);
	unsigned int seq_units = num_base/DNA2UnitSizeInBase;
	seq_remainder = num_base % DNA2UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return (*this);
	}
	
	seq = new unsigned char[len];
	
	// Only need to initialize the last byte
	seq[len - 1] = DEFAULT_BYTE;
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq;
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr += DNA2UnitSizeInBase, head++ /*head += DNA2UnitSizeInByte*/){
				
		WriteUnit2(head, ptr);
	}
	
	WriteUnit2Remainder(head, ptr, seq_remainder);
		
	return (*this);
}

DNA2Seq DNA2Seq::operator+(const DNA2Seq &m_seq)
{	
	DNA2Seq tmp(this->size() + m_seq.size());
	
	DNA2Seq::iterator iter = tmp.begin();
	
	DNA2Seq::iterator copy_iter;
	
	for(copy_iter = this->begin();copy_iter != this->end();copy_iter++, iter++){
		iter.base( copy_iter.base() );
	}
	
	for(copy_iter = m_seq.begin();copy_iter != m_seq.end();copy_iter++, iter++){
		iter.base( copy_iter.base() );
	}
	
	return tmp;
}

DNA2Seq DNA2Seq::subseq(const int &m_start, const unsigned int &m_len,
	const bool &m_circular /* = false */) const
{

	iterator iter;
	
	if(m_circular){
		if(m_start > 0){
			iter = begin() + m_start % size();
		}
		else{
			iter = begin() + (size() - m_start%size());
		}
	}
	else{
		if((m_start + m_len  - 1) >= size()){
			throw __FILE__ ":DNA2Seq::subseq: Illegal range";
		}
		
		iter = begin() + m_start;
	}
	
	DNA2Seq tmp(m_len);
	iterator tmp_iter = tmp.begin();
	
	// This was the original version that did not handle periodic sequences
	//iterator iter = begin() + m_start;
	
	for(unsigned int i = 0;i < m_len;i++, iter++, tmp_iter++){
		if(m_circular && (iter == end())){
			iter = begin();
		}
		
		tmp_iter.base(iter.base());
	}
	
	return tmp;
}

// Make a copy of the complement of the given sequence
// where complement is the reversed sequence with following
// base substitutions:
// A->T'
// T->A'
// C->G'
// G->C'
void DNA2Seq::complement(const char *m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = strlen(m_seq);
	unsigned int seq_units = num_base/DNA2UnitSizeInBase;
	seq_remainder = num_base % DNA2UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len];
	
	// Only need to initialize the last byte
	seq[len - 1] = DEFAULT_BYTE;
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq + num_base - 1;
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr -= DNA2UnitSizeInBase, head++ /*head += DNA2UnitSizeInByte*/){
				
		WriteUnit2Complement(head, ptr);
	}
	
	WriteUnit2ComplementRemainder(head, ptr, seq_remainder);
}

// Make a copy of the complement of the given sequence
// where complement is the reversed sequence with following
// base substitutions:
// A->T'
// T->A'
// C->G'
// G->C'
void DNA2Seq::complement(const string &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = m_seq.size();
	unsigned int seq_units = num_base/DNA2UnitSizeInBase;
	seq_remainder = num_base % DNA2UnitSizeInBase;
	
	len = seq_units + (seq_remainder != 0);
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len];
	
	// Only need to initialize the last byte
	seq[len - 1] = DEFAULT_BYTE;
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	unsigned char *ptr = (unsigned char*)m_seq.c_str() + num_base - 1;
	unsigned char *head = seq;
	
	for(unsigned int i = 0;i < seq_units;i++, 
		ptr -= DNA2UnitSizeInBase, head++/*head += DNA2UnitSizeInByte*/){
				
		WriteUnit2Complement(head, ptr);
	}
	
	WriteUnit2ComplementRemainder(head, ptr, seq_remainder);
}

void DNA2Seq::complement(const DNA2Seq &m_seq)
{
	if(seq){
		delete [] seq;
	}
	
	// How much space will be required to store the sequence?
	num_base = m_seq.num_base;
	len = m_seq.len;
	seq_remainder = m_seq.seq_remainder;
	
	if(len == 0){
		seq = NULL;
		return;
	}
	
	seq = new unsigned char[len];
	
	// Only need to initialize the last byte
	seq[len - 1] = DEFAULT_BYTE;
	
	#ifdef SEQ_DEBUG
	if(seq == NULL){
		throw __FILE__ ":DNA2Seq: Unable to allocate memory";
	}
	#endif // SEQ_DEBUG
	
	complement_iterator comp_iter = m_seq.complement_begin();
	iterator iter = begin();

	while(comp_iter != m_seq.complement_end()){
		iter.base(comp_iter.base());
		
		iter++;
		comp_iter++;
	}
}

DNA2Seq DNA2Seq::complement() const
{
	DNA2Seq tmp;

	tmp.complement(*this);

	return tmp;
}

// Convert this DNA2Seq into a text string
string  DNA2Seq::str() const
{
	string tmp;
	int i;
	
	tmp.resize(num_base);
	
	DNA2Seq::iterator iter;
	
	for(i = 0, iter =  begin();iter !=  end();iter++, i++){
		
		tmp[i] = iter.c_base();
	}
	
	return tmp;
}
	
ostream& operator << (ostream &s, const DNA2Seq &m_seq)
{
	DNA2Seq::iterator iter;
	unsigned int count = 0;
	
	for(iter =  m_seq.begin();iter !=  m_seq.end();iter++){
		
		s << iter.c_base();
		
		if(wrap_seq_output){
			count++;

			if(count % DNA_OUTPUT_LINE_LENGTH == 0){
				s << endl;
			}
		}
	}
	
	return s;
}

inline void WriteUnit2(unsigned char *unit, unsigned char *ptr)
{
	*unit = 	(CharToBase2(ptr[0]) << 6) + 
			(CharToBase2(ptr[1]) << 4) + 
			(CharToBase2(ptr[2]) << 2) + 
			CharToBase2(ptr[3]);
}

inline void WriteUnit2Remainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder)
{
	switch(remainder){
		case 1:
			*unit = (CharToBase2(ptr[0]) << 6);
			return;
		case 2:
			*unit = (CharToBase2(ptr[0]) << 6) + 
					(CharToBase2(ptr[1]) << 4);
			return;
		case 3:
			*unit = (CharToBase2(ptr[0]) << 6) + 
					(CharToBase2(ptr[1]) << 4) + 
					(CharToBase2(ptr[2]) << 2);
			
			return;
	};
}

inline void WriteUnit2Complement(unsigned char *unit, unsigned char *ptr)
{
	*unit = 	(CharToComplementBase2(ptr[0]) << 6) + 
			(CharToComplementBase2(ptr[-1]) << 4) + 
			(CharToComplementBase2(ptr[-2]) << 2) + 
			CharToComplementBase2(ptr[-3]);
}

inline void WriteUnit2ComplementRemainder(unsigned char *unit, unsigned char *ptr, 
	const unsigned int &remainder)
{
	switch(remainder){
		case 1:
			*unit = (CharToComplementBase2(ptr[0]) << 6);
			return;
		case 2:
			*unit = (CharToComplementBase2(ptr[0]) << 6) + 
					(CharToComplementBase2(ptr[-1]) << 4);
			return;
		case 3:
			*unit = (CharToComplementBase2(ptr[0]) << 6) + 
					(CharToComplementBase2(ptr[-1]) << 4) + 
					(CharToComplementBase2(ptr[-2]) << 2);
			
			return;
	};
}

inline unsigned char CharToBase2(const unsigned char &c)
{
	switch(c){
		case 'a':
		case 'A':
			return DNA2Seq::A;
		case 't':
		case 'T':
			return DNA2Seq::T;
		case 'g':
		case 'G':
			return DNA2Seq::G;
		case 'c':
		case 'C':
			return DNA2Seq::C;
		default:
			throw __FILE__ ":CharToBase2: Unknown base (not A, G, C or T)";
	};
	
	return 0;
}

inline unsigned char CharToComplementBase2(const unsigned char &c)
{
	switch(c){
		case 'a':
		case 'A':
			return DNA2Seq::T;
		case 't':
		case 'T':
			return DNA2Seq::A;
		case 'g':
		case 'G':
			return DNA2Seq::C;
		case 'c':
		case 'C':
			return DNA2Seq::G;
		default:
			throw __FILE__ ":CharToComplementBase2: Unknown base (not A, G, C or T)";
	};
	
	return 0;
}

inline unsigned char BaseToChar2(const unsigned char &c)
{	
	switch(c & BITMASK2){
		case DNA2Seq::A:
			return 'A';
		case DNA2Seq::T:
			return 'T';
		case  DNA2Seq::G:
			return 'G';
		case DNA2Seq::C:
			return 'C';
	};
	
	return 0;
}
